# -*- coding: utf-8 -*-

import json
import logging.config
import re


class RelpathFilter(logging.Filter):
    """
    This filter adds a relpath contextual information to logging records.

    The relative path is computed with respect to the regular expressions given
    to the constructor (replaced with a blank string).
    """

    def __init__(self, regexps=None):
        """
        """
        if regexps is None:
            regexps = []
        self.regexps = regexps

    def filter(self, record):
        """
        """
        record.relpath = record.pathname
        for regexp in self.regexps:
            record.relpath = re.sub(regexp, "", record.relpath)

        return True


class RelsourceFilter(RelpathFilter):
    """
    This filter adds a relsource (relpath:lineno) contextual information to
    logging records.

    The relative path is computed with respect to the regular expressions given
    to the constructor (replaced with a blank string).
    """

    def __init__(self, regexps=None):
        """
        """
        super().__init__(regexps)

    def filter(self, record):
        """
        """
        result = super().filter(record)
        record.relsource = record.relpath + ":" + str(record.lineno)

        return result


class logger(object):
    """
    """

    @staticmethod
    def configure(filename):
        """
        """
        with open(filename) as f:
            config = json.load(f)
            logging.config.dictConfig(config)

        logger = logging.getLogger("moustache_fusion")
        logger.debug("Using logging config file \"%s\"" % filename)

    @staticmethod
    def __getattr__(name):
        """
        """
        return getattr(logging.getLogger("moustache"), name)



# -*- coding: utf-8 -*-

import pytest

from moustache_fusion import webservice


@pytest.fixture
def client():
    webservice.app.config['TESTING'] = True

    with webservice.app.test_client() as client:
        yield client



# -*- coding: utf-8 -*-

import subprocess
from PyPDF2 import PdfFileWriter, PdfFileReader
import io
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import A4
from itertools import count, product
from string import ascii_uppercase
import tempfile

from moustache_fusion.logger import logger
from moustache_fusion.exceptions import InvalidUsage

PDFTK_PATH = '/usr/bin/pdftk'
PDFINFO_PATH = '/usr/bin/pdfinfo'
PDFGREP_PATH = '/usr/bin/pdfgrep'


# alphabet de lettre (A,B, ..., AB, AC)
def multiletters(seq):
    for n in count(1):
        for s in product(seq, repeat=n):
            yield ''.join(s)


letter = multiletters(ascii_uppercase)


def set_annexes_length(annexes):
    for annexe in annexes:
        logger().debug("exec %s %s" % (PDFINFO_PATH, annexe['path']))
        output = subprocess.run([PDFINFO_PATH, annexe['path']], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if output.returncode:
            raise Exception(output.stderr.decode('utf-8'))
        output_array = output.stdout.decode('utf-8').split()
        pageinfo = output_array.index("Pages:")
        v = output_array[pageinfo + 1]
        if not v.isdigit():
            raise Exception("Impossible de lire le nombre de pages dans %s:%s" % (annexe['path'], v))
        annexe['length'] = int(v)
        logger().debug("%s pages:%d" % (annexe['path'], annexe['length']))


# find only one + cache file speed +++
# pdfgrep -n --cache "challenge" principal.pdf
# find pattern page
# TODO test if not null
def set_annexe_pattern_pages(annexes, document_path):
    new_annexes = []
    for annex in annexes:
        # ubuntu 16.04 ships with pdfgrep 1.4 which does not support '--cache' parameter
        # cmd = '{0} -n --cache "{1}" {2}'.format(PDFGREP_PATH, annex['pattern'], documentPath)
        cmd = '{0} -n --cache "{1}" {2}'.format(PDFGREP_PATH, annex['pattern'], document_path)
        logger().debug("exec %s" % cmd)
        res = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        outds, errs = res.communicate()
        if res.returncode:
            if not errs:
                errs = "%s retourne %d" % (PDFGREP_PATH, res.returncode)
            raise Exception(errs)

        outds = outds.split()
        for line in outds:
            v = line.split(":")[0]
            if not v.isdigit():
                raise Exception("Impossible de lire la page de départ pour %s" % annex['name'])
            annex['alias'] = next(letter)
            annex['startPage'] = int(v)
            new_annexes.append(dict(annex))
            logger().debug("%s alias=%s start page=%d" % (annex['name'], annex['alias'], annex['startPage']))
    return new_annexes


# create and execute script to replace annexes
# cmd = "/usr/bin/pdftk A=main.pdf B=annexe.pdf cat A1-10 B A12-end output generated.pdf"
# TODO test if not null
def replace_annexes(docs, outputfile):
    # generate aliases (/usr/bin/pdftk A=main.pdf B=annexe.pdf)
    # sort Annexes by position
    docs['annexes'].sort(key=lambda ann: ann['startPage'])

    docs['general']['alias'] = next(letter)
    cmd = "{0} {1}={2} ".format(PDFTK_PATH, docs['general']['alias'], docs['general']['path'])
    logger().debug(docs)
    for annex in docs['annexes']:
        annex['alias'] = next(letter)
        # cmd += annex['alias'] + "=" + annex['path'] + " "
        cmd += "{0}={1} ".format(annex['alias'], annex['path'])

    # (cat A1-10 B A12-end output generated.pdf)
    cmd += "cat "
    general_doc_position = 1

    for annex in docs['annexes']:
        if general_doc_position <= annex['startPage'] - 1:
            cmd += "{0}{1}-{2} {3} ".format(
                docs['general']['alias'],
                general_doc_position,
                annex['startPage'] - 1,
                annex['alias']
            )
        else:
            cmd += "{0} ".format(annex['alias'])

        general_doc_position = annex['startPage'] + annex['length']

    cmd += "{0}{1}-end output {2}".format(docs['general']['alias'], general_doc_position, outputfile)

    logger().debug(cmd)
    res = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    outds, errs = res.communicate()
    if res.returncode:
        raise Exception(errs)


def add_page_numbering_to_annexes(docs):
    for annexe in docs['annexes']:
        start = annexe['startPage']
        existing_pdf = PdfFileReader(open(annexe['path'], "rb"))
        output = PdfFileWriter()
        end = start + annexe['length']
        increment = 0
        while start < end:
            packet = io.BytesIO()
            # @todo: parameters
            le_can = canvas.Canvas(packet, pagesize=A4)
            le_can.setFont('Helvetica', 10)
            le_can.drawString(200, 20, "{0}".format(start))
            le_can.save()

            # move to the beginning of the buffer
            packet.seek(0)
            new_pdf = PdfFileReader(packet)

            logger().debug("%s cur=%d [%d-%d]" % (annexe['path'], increment, start, end))
            page = existing_pdf.getPage(increment)
            page.mergePage(new_pdf.getPage(0))
            output.addPage(page)
            increment += 1
            start += 1

        # finally, write "output" to a real file
        new_annexe_name = tempfile.NamedTemporaryFile(suffix='.pdf', delete=False).name
        output_stream = open(new_annexe_name, 'wb')
        output.write(output_stream)
        output_stream.close()
        logger().debug("%s to %s" % (annexe['path'], new_annexe_name))
        annexe['path'] = new_annexe_name


def check_patterns(data: dict) -> None:
    """
    Check that all patterns are unique or raise an exception.
    """
    found = {}
    for pdf in data['annexes']:
        if pdf['pattern'] in found:
            msgstr = 'Non unique pattern "%s" found for file "%s" (already found for file "%s")'
            raise InvalidUsage(msgstr % (pdf['pattern'], pdf['name'], found[pdf['pattern']]))
        found[pdf['pattern']] = pdf['name']


# @todo: temp_dir parameter, return file path
def skittles(data: dict, outputfile: str) -> None:
    logger().debug("skittles starts")

    check_patterns(data)

    set_annexes_length(data['annexes'])
    logger().debug('************** %s' % data)
    data['annexes'] = set_annexe_pattern_pages(data['annexes'], data['general']['path'])
    logger().debug('************** %s' % data)

    try:
        with_page_numbering = data["options"]["with_annexes_pages_numbered"]
    except KeyError:
        with_page_numbering = False

    if with_page_numbering:
        add_page_numbering_to_annexes(data)
    else:
        logger().debug("skip add_page_numbering_to_annexes")
    replace_annexes(data, outputfile)

    logger().debug("file created: %s" % outputfile)



# -*- coding: utf-8 -*-

import os
import shutil

from PIL import Image
from pdf2image import convert_from_path
from pixelmatch.contrib.PIL import pixelmatch


def compare_images(reference_path: str, candidate_path: str, diff_path: str) -> int:
    img_a = Image.open(reference_path)
    img_b = Image.open(candidate_path)
    img_diff = Image.new('RGB', img_a.size)

    diffs = pixelmatch(img_a, img_b, img_diff, includeAA=True)
    img_diff.save(diff_path)

    return diffs


def list_files(dir: str) -> list:
    return sorted([file for file in os.listdir(dir) if os.path.isfile(os.path.join(dir, file))])


def compare_exported_images(reference_path: str, candidate_path: str, diffs_path) -> dict:
    # List dirs contents
    references = list_files(reference_path)
    candidates = list_files(candidate_path)

    if os.path.exists(diffs_path):
        shutil.rmtree(diffs_path)
    os.makedirs(diffs_path)

    result = {
        'missing': sorted(list(set(references) - set(candidates))),
        'extra': sorted(list(set(candidates) - set(references))),
        'common': {key: None for key in sorted(list(set(candidates) & set(references)))}
    }

    # for each common page, make diffs
    for basename in result['common'].keys():
        result['common'][basename] = compare_images(
            reference_path + '/' + basename,
            candidate_path + '/' + basename,
            diffs_path + '/' + basename
        )

    return result


def export_pdf_to_images(pdf_path: str, output_folder: str) -> None:
    if os.path.exists(output_folder):
        shutil.rmtree(output_folder)
    os.makedirs(output_folder)

    tmp_paths = convert_from_path(
        pdf_path,
        dpi=100,
        fmt='jpeg',
        output_folder=output_folder,
        paths_only=True,
        use_pdftocairo=True
    )

    format = '/{:05d}.jpg'
    for idx, src_path in enumerate(tmp_paths, start=1):
        dst_path = output_folder + format.format(idx)
        shutil.move(src_path, dst_path)



# -*- coding: utf-8 -*-

def get_project_dir() -> str:
    """
    Returns absolute project dir.
    """
    return '/app'


def get_fixture_path(path: str) -> str:
    """
    Returns absolute fixture file path.

    :param path: Relative path to the file
    """
    return get_project_dir() + "/tests/fixtures/" + path



# -*- coding: utf-8 -*-

class InvalidUsage(Exception):
    status_code = 400

    def __init__(self, message, status_code=400, payload=None):
        Exception.__init__(self)
        self.message = message
        self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv

# @todo: CommandException



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
curl -i -X POST http://localhost:5000/api \
    -F "params=@ressources/test-ok/skittles.json" \
    -F "principal=@ressources/test-ok/principal.pdf" \
    -F "annexe=@ressources/test-ok/annexe1.pdf" \
    -F "annexe=@ressources/test-ok/annexe2.pdf" > resultws.pdf
"""
import json
import magic
import pathlib
import tempfile
import os
import getopt
import sys
import shutil
import logging

from flask import Flask, request, send_file, jsonify, render_template
from io import StringIO
from logging.config import fileConfig
from werkzeug.exceptions import InternalServerError

from moustache_fusion import skittlespy
from moustache_fusion.datatransfer import SwapRequest
from moustache_fusion.exceptions import InvalidUsage
from moustache_fusion.logger import logger

logger().configure('/app/config/logging_config.json')


class API:
    class V1:
        class PDF:
            class SWAP:
                ANNEXES = 'annexe'
                FILE = 'principal'
                JSON = 'params'
                #  @todo: /api/v1/pdf/swap
                URL = '/api'


#  @todo: report enhancements in moustache
#  @todo: raise error if key already exists in retrieve_multiple
#  @todo: "mandatory" kwarg for both functions (default True)


class FileRetriever():
    """
    This static class provides static utility methods to retrieve files from the flask (werkzeug) HTTP request by key.
    """

    @classmethod
    def _dest_file(cls, filename: str, dest_directory: str, *, temp_name: bool = False):
        if temp_name is False:
            dest_file = os.path.join(dest_directory, filename)
        else:
            suffix = ''.join(pathlib.Path(filename).suffixes)
            dest_file = tempfile.NamedTemporaryFile(dir=dest_directory, delete=False, suffix=suffix).name
        return dest_file

    @classmethod
    def retrieve_single(cls, key: str, dest_directory: str, *, temp_name: bool = False):
        """
        Retrieves a single (mandatory) file from the request by key and stores it in the destination directory,
        returning the complete path to the saved file.
        """
        if key not in request.files:
            raise InvalidUsage("\"%s\" key is not in request files (%s)" % (key, request.files))

        file = request.files[key]
        if file.filename.strip() == '':
            raise InvalidUsage("\"%s\" key has an empty filename in request files (%s)" % (key, request.files))

        dest_file = cls._dest_file(file.filename, dest_directory, temp_name=temp_name)

        logger().debug("Retrieving \"%s\" file \"%s\" to \"%s\"" % (key, file.filename, dest_file))
        file.save(dest_file)
        return dest_file

    @classmethod
    def retrieve_multiple(cls, key: str, dest_directory: str, *, temp_name: bool = False):
        """
        Retrieves a (possibly empty) list of files from the request by key (possibly non-existent) and stores them in
        the destination directory, returning a dict with original filenames as key and complete path to the saved file
        as value.
        """
        filelist = request.files.getlist(key)
        file_mapping = {}
        logger().debug("Retrieving %d %s file(s)..." % (len(filelist), key))

        for index, file in enumerate(filelist):
            if file.filename.strip() == '':
                msgstr = "\"%s\"[%d] key has an empty filename in request files (%s)"
                raise InvalidUsage(msgstr % (key, index + 1, request.files))

            dest_file = cls._dest_file(file.filename, dest_directory, temp_name=temp_name)
            logger().debug(
                "... retrieving %s %d \"%s\" (%s) to \"%s\"" % (
                    key,
                    index + 1,
                    file.filename,
                    'None' if file.mimetype is None or file.mimetype == '' else "'" + file.mimetype + '"',
                    dest_file
                )
            )
            file.save(dest_file)
            file_mapping[file.filename] = dest_file

        return file_mapping


app = Flask(__name__)
# TODO a virer ?
app.secret_key = 'super secret key'


def setlogger(conffile):
    if not os.path.isfile(conffile):
        logging.getLogger().error("Can't access %s" % conffile)
        sys.exit(1)

    fileConfig(conffile)
    logger = logging.getLogger()
    logger.debug("Using %s for logging config file" % conffile)
    logger().handlers = logger.handlers


@app.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    logger().debug('... API: invalid usage, aborting request (%s)' % (error.message))
    return response


@app.route('/')
def www_v1_index():
    logger().debug('WWW: rendering HTML template')
    return render_template('index.html', API=API.V1)


# @todo: Check if json is compliant
@app.route(API.V1.PDF.SWAP.URL, methods=['POST'])
def api_v1_pdf_swap():
    logger().debug('API: starting swap...')

    temp_directory = None

    try:
        swap_request = SwapRequest()
        temp_directory = tempfile.mkdtemp()

        # @todo: test with empty / unreadable JSON
        # @todo: add json/json_stream to FileRetriver
        if API.V1.PDF.SWAP.JSON not in request.files:
            raise InvalidUsage("Le fichier json de paramètres n'est pas présent")

        j = request.files['params']
        json_content = j.stream.read()
        data = json.load(StringIO(json_content.decode('utf-8')))
        logger().debug("... retrieved JSON data %s" % (data))

        #  @todo: with_annexes_pages_numbered should come from the "form"
        swap_request.main_pdf = FileRetriever.retrieve_single(API.V1.PDF.SWAP.FILE, temp_directory, temp_name=True)
        mime = magic.from_file(swap_request.main_pdf, mime=True)
        if mime != 'application/pdf':
            msgstr = 'Main PDF document "%s" MIME type is not "application/pdf" (found "%s" MIME type for "%s")'
            raise InvalidUsage(msgstr % (API.V1.PDF.SWAP.FILE, mime, request.files[API.V1.PDF.SWAP.FILE].filename))

        # @todo: DTO
        data['general'] = {}
        data['general']['path'] = swap_request.main_pdf

        annexes_filelist = FileRetriever.retrieve_multiple(API.V1.PDF.SWAP.ANNEXES, temp_directory, temp_name=True)
        # @todo deal with extra files in annexes_filelist ?
        # @todo pb si nom de fichier non unique ?
        for annexe in data['annexes']:
            if annexe['name'] in annexes_filelist:
                # @todo: annexe['path'], and keep name
                pdf = {'name': annexe['name'], 'path': annexes_filelist[annexe['name']], 'pattern': annexe['pattern']}
                swap_request.pdfs.append(pdf)

                annexe['path'] = annexes_filelist[annexe['name']]

                mime = magic.from_file(pdf['path'], mime=True)
                if mime != 'application/pdf':
                    msgstr = 'Document "%s" MIME type is not "application/pdf" (found "%s" MIME type for "%s")'
                    raise InvalidUsage(msgstr % (API.V1.PDF.SWAP.ANNEXES, mime, pdf['name']))
            else:
                msgstr = 'File name "%s" present in "%s" JSON data but not in "%s" files'
                raise InvalidUsage(msgstr % (annexe['name'], API.V1.PDF.SWAP.JSON, API.V1.PDF.SWAP.ANNEXES))

        output_file = tempfile.NamedTemporaryFile(dir=temp_directory, delete=False, suffix='.pdf').name

        #  @todo: service
        logger().debug("============ swap_request=%s output_file=%s" % (swap_request, output_file))
        logger().debug("============ data=%s output_file=%s" % (data, output_file))

        skittlespy.skittles(data, output_file)
        result = send_file(output_file, attachment_filename='result.pdf', as_attachment=True)
    except InvalidUsage:
        raise
    except Exception as exc:
        logger().exception(exc)
        raise InternalServerError(str(exc))
    finally:
        if temp_directory is not None:
            shutil.rmtree(temp_directory)

    logger().debug('API: ...swap succeeded')
    return result


def default_app():
    return app


def usage():
    print("usage :")
    print("-p --port=port\tport d'écoute")
    print("-d --debug\t\tactive les traces sur stderr")
    print("-l --logger=loggerfile\t\tfichier de configuration du logger")


if __name__ == '__main__':
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hp:dl:", ["help", "port=", "debug", "logger="])
    except getopt.GetoptError as err:
        print(err)
        usage()
        sys.exit(1)

    portparam = None
    debugparam = False
    loggerparam = None

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit(0)
        elif o in ("-p", "--port"):
            portparam = int(a)
        elif o in ("-d", "--debug"):
            debugparam = True
        elif o in ("-l", "--logger"):
            loggerparam = a
        else:
            print("unhandled option")
            usage()
            sys.exit(1)

    if loggerparam:
        setlogger(loggerparam)

    app.run(debug=debugparam, host='0.0.0.0', port=portparam, threaded=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import getopt
import sys
import skittlespy
import json
import os
import logging
from logging.config import fileConfig

logger = logging.getLogger()


def usage():
    print("usage :")
    print("-i --in=pdfinfile\tfichier pdf d'entrée")
    print("-c --conf=conffile\tfichier de paramètres")
    print("-o --out=pdfoutfile\tfichier pdf de sortie")
    print("-d --debug\t\tactive les traces sur stderr")
    print("-l --logger=loggerfile\t\tfichier de configuration du logger")


def normalize_parameters(parameters, infile, filepath):
    filepath = os.path.realpath(filepath)
    dirname = os.path.dirname(filepath)

    if not os.path.isabs(infile):
        # pdf en entrée : si relatif par rapport au cwd
        infile = os.path.realpath(infile)
    parameters["general"] = dict()
    parameters["general"]["name"] = infile

    for annexe in parameters["annexes"]:
        if not os.path.isabs(annexe["name"]):
            # les annexes en relatif par rapport au fichier de config
            annexe["name"] = os.path.realpath(os.path.join(dirname, annexe["name"]))

    return parameters


def setlogger(conffile):
    if not os.path.isfile(conffile):
        logger.error("Can't access %s" % conffile)
        sys.exit(1)

    fileConfig(conffile)
    logger.debug("Using %s for logging config file" % conffile)


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hi:c:o:dl=", ["help", "in=", "conf=", "out=", "debug", "logger="])
    except getopt.GetoptError as err:
        print(err)
        usage()
        sys.exit(1)

    inparam = None
    confparam = None
    outparam = None
    debugparam = False
    loggerparam = None

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit(0)
        elif o in ("-i", "--in"):
            inparam = a
        elif o in ("-c", "--conf"):
            confparam = a
        elif o in ("-o", "--out"):
            outparam = a
        elif o in ("-d", "--debug"):
            debugparam = True
        elif o in ("-l", "--logger"):
            loggerparam = a
        else:
            print("unhandled option")
            usage()
            sys.exit(1)

    if loggerparam:
        setlogger(loggerparam)

    if inparam is None or confparam is None or outparam is None:
        usage()
        sys.exit(1)

    if not os.path.isfile(inparam):
        logger.error("%s not found" % inparam)
        sys.exit(1)

    if not os.path.isfile(confparam):
        logger.error("%s not found" % confparam)
        sys.exit(1)
    try:
        with open(confparam) as json_data:
            parameters = json.load(json_data)
    except Exception as e:
        logger.error("%s invalid format\n%s" % (confparam, e))
        sys.exit(1)

    parameters = normalize_parameters(parameters, inparam, confparam)

    res = skittlespy.skittles(parameters, outparam, debugparam)
    if res is not True:
        logger.error(res)
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import moustache_fusion
import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='Moustache-fusion.beta',
    version=moustache_fusion.__version__,
    python_requires='>=3',
    packages=setuptools.find_packages(),
    author='Libriciel SCOP',
    author_email='hackathon@libriciel.coop',
    description='Module post-Moustache pour fusion d\'annexes PDF dans PDF principal',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'flask>=1.1,<1.2',
        'python-magic>=0.4,<0.5',
        'PyPDF2>=1.26,<1.27',
        'reportlab>=3.5,<3.6'
    ],
    extras_require={
        'dev': [
            'coverage>=5.3,<5.4',
            'flake8>=3.8,<3.9',
            'pdf2image>=1.14,<1.15',
            'pixelmatch>=0.2,<0.3',
            'pytest>=6.1,<6.2'
        ]
    },
    include_package_data=True,
    url='https://gitlab.libriciel.fr/outils/skittlespy/',
    entry_points={
        'console_scripts': [
            'moustache_fusion = moustache_fusion:launch'
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6'
    ],
    license='CeCILL-2.1',
)



# -*- coding: utf-8 -*-

"""
    Module de fusion documentaire
"""
__version__ = '0.0.1-162969'


def launch():
    from moustache_fusion import webservice
    webservice.default_app().run(debug=True, host='0.0.0.0')
